#include "DirectX9App.h"
#include "DirectInput.h"
#include "DirectX9Game.h"
#include <ctime>
#include "AssetManager.h"
#include "Messenger.h"
#include "FMODAudio.h"

//Contains the main game loop. Initializes and shuts down all other systems.
int DirectX9App::Run()
{
	MSG  msg;
    msg.message = WM_NULL;

	__int64 cntsPerSec = 0;
	QueryPerformanceFrequency((LARGE_INTEGER*)&cntsPerSec);
	float secsPerCnt = 1.0f / (float)cntsPerSec;

	__int64 prevTimeStamp = 0;
	QueryPerformanceCounter((LARGE_INTEGER*)&prevTimeStamp);
		
	//Random number gen seed
	while(msg.message != WM_QUIT)
	{
		// If there are Window messages then process them.
		if(PeekMessage( &msg, 0, 0, 0, PM_REMOVE ))
		{
            TranslateMessage( &msg );
            DispatchMessage( &msg );
		}
		else
        {	
			// If the application is paused then free some CPU cycles to other 
			// applications and then continue on to the next frame.
			if( m_AppPaused )
			{
				Sleep(20);
				continue;
			}

			//Main Game Loop.
			if( !isDeviceLost() )
			{
				__int64 currTimeStamp = 0;
				QueryPerformanceCounter((LARGE_INTEGER*)&currTimeStamp);
				float dt = (currTimeStamp - prevTimeStamp)*secsPerCnt;
				//Update primary systems here
				gInput->Update();
				gMessenger->Update(dt);
				gAudio->Update();
				//Set Exit command
				if(gInput->KeyDown(DIK_ESCAPE))
				{
					PostQuitMessage(0);
				}
				//Update Game
				gGame->Update(dt);
				//Update window rect
				m_WindowRect.top = 0;
				m_WindowRect.left = 0;
				m_WindowRect.right = (LONG)GetWindowSize().x;
				m_WindowRect.bottom = (LONG)GetWindowSize().y;
				//Begin rendering
				BeginFrame();
				//Render Input related objects (cursor and/or mouse position)
				//gInput->Render();
				//Render Game
				gGame->Render();
				//End rendering
				EndFrame();
				prevTimeStamp = currTimeStamp;
			}
        }
    }
	//Shutdown primary systems
	gInput->Shutdown();
	return (int)msg.wParam;
}

//Release non mananged objects
void DirectX9App::onLostDevice()
{
	m_SpriteInterface->OnLostDevice();
	gAssets->onLostDevice();
}

//Reset non managed objects
void DirectX9App::onResetDevice()
{
	m_SpriteInterface->OnResetDevice();
	gAssets->onResetDevice();
}

//Link to App instance
DirectX9App* DirectX9App::Instance()
{
	static DirectX9App instance;
	return &instance;
}

DirectX9App::DirectX9App():
	m_MainWndCaption(APP_TITLE),
	m_hMainWnd(0),
	m_D3DObject(0),
	m_D3DDevice(0),
	m_ClearColor(1.0f,1.0f,1.0f,1.0f),		//Blue clear color
	m_AppPaused(false),
	m_SpriteInterface(0)
{
	//Data Member Initialization
	ZeroMemory(&m_D3DPP, sizeof(m_D3DPP));
	//Init Window
	InitMainWindow();
	//Init DirectX
	InitDirect3D();
	//Seed random numbers
	srand(time(NULL));
	//Load background
}

DirectX9App::~DirectX9App()
{
	Shutdown();
}

void DirectX9App::EnableFullScreenMode(bool enable)
{
	// Switch to fullscreen mode.
	if( enable )
	{
		// Are we already in fullscreen mode?
		if( !m_D3DPP.Windowed ) 
			return;

		//Get screen size
		int width  = GetSystemMetrics(SM_CXSCREEN);
		int height = GetSystemMetrics(SM_CYSCREEN);

		//Reset backbuffer
		m_D3DPP.BackBufferFormat = D3DFMT_X8R8G8B8;
		m_D3DPP.BackBufferWidth  = width;
		m_D3DPP.BackBufferHeight = height;
		m_D3DPP.Windowed         = false;

		// Change the window style to a more fullscreen friendly style.
		SetWindowLongPtr(m_hMainWnd, GWL_STYLE, WS_POPUP);
		SetWindowPos(m_hMainWnd, HWND_TOP, 0, 0, width, height, SWP_NOZORDER | SWP_SHOWWINDOW);	
	}
	// Switch to windowed mode.
	else
	{
		// Are we already in windowed mode?
		if( m_D3DPP.Windowed ) 
			return;

		//Reset Backbuffer
		RECT R = {0, 0, BASE_WINDOW_WIDTH, BASE_WINDOW_HEIGHT};
		AdjustWindowRect(&R, WS_OVERLAPPEDWINDOW, false);
		m_D3DPP.BackBufferFormat = D3DFMT_UNKNOWN;
		m_D3DPP.BackBufferWidth  = 1024;
		m_D3DPP.BackBufferHeight = 512;
		m_D3DPP.Windowed         = true;
	
		// Change the window style to a more windowed friendly style.
		SetWindowLongPtr(m_hMainWnd, GWL_STYLE, WS_OVERLAPPEDWINDOW);
		SetWindowPos(m_hMainWnd, HWND_TOP, 100, 100, R.right, R.bottom, SWP_NOZORDER | SWP_SHOWWINDOW);
	}


				

	// Reset the device with the changes.
	onLostDevice();
	HR(m_D3DDevice->Reset(&m_D3DPP));
	onResetDevice();
}

bool DirectX9App::isDeviceLost()
{
	// Get the state of the graphics device.
	HRESULT hr = gD3DDevice->TestCooperativeLevel();

	// If the device is lost and cannot be reset yet then
	// sleep for a bit and we'll try again on the next 
	// message loop cycle.
	if( hr == D3DERR_DEVICELOST )
	{
		Sleep(20);
		return true;
	}
	// Driver error, exit.
	else if( hr == D3DERR_DRIVERINTERNALERROR )
	{
		MessageBox(0, "Internal Driver Error...Exiting", 0, 0);
		PostQuitMessage(0);
		return true;
	}
	// The device is lost but we can reset and restore it.
	else if( hr == D3DERR_DEVICENOTRESET )
	{
		onLostDevice();
		HR(gD3DDevice->Reset(&m_D3DPP));
		onResetDevice();
		return false;
	}
	else
		return false;
}

bool DirectX9App::CheckDeviceCaps()
{
	D3DCAPS9 caps;
	HR(m_D3DDevice->GetDeviceCaps(&caps));

	// Check for vertex shader version 2.0 support.
	if( caps.VertexShaderVersion < D3DVS_VERSION(2, 0) )
		return false;

	// Check for pixel shader version 2.0 support.
	if( caps.PixelShaderVersion < D3DPS_VERSION(2, 0) )
		return false;

	// Check render target support.
	D3DDISPLAYMODE mode;
	HR(m_D3DObject->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &mode));

	// Windowed.
	if(FAILED(m_D3DObject->CheckDeviceFormat(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, mode.Format, 
		D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_R32F)))
		return false;
	if(FAILED(m_D3DObject->CheckDepthStencilMatch(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, mode.Format,
		D3DFMT_R32F, D3DFMT_D24X8)))
		return false;

	// Fullscreen.
	if(FAILED(m_D3DObject->CheckDeviceFormat(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, D3DFMT_X8R8G8B8, 
		D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_R32F)))
		return false;
	if(FAILED(m_D3DObject->CheckDepthStencilMatch(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, D3DFMT_X8R8G8B8,
		D3DFMT_R32F, D3DFMT_D24X8)))
		return false;
	return true;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	//If DirectX has not be initialized, return the default window proc.
	if(gD3DDevice != 0)
		return DirectX9App::Instance()->MessageProc(msg, wParam, lParam);
	else
		return DefWindowProc(hWnd, msg, wParam, lParam);
}

void DirectX9App::InitMainWindow()
{
	WNDCLASS wc;
	wc.style         = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc   = WndProc;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = m_hAppInst;
	wc.hIcon         = LoadIcon(0, IDI_APPLICATION);
	wc.hCursor       = LoadCursor(0, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName  = 0;
	wc.lpszClassName = "D3DWndClassName";

	if( !RegisterClass(&wc) )
	{
		MessageBox(0, "RegisterClass FAILED", 0, 0);
		PostQuitMessage(0);
	}

	RECT R = {0, 0, BASE_WINDOW_WIDTH, BASE_WINDOW_HEIGHT};
	AdjustWindowRect(&R, WS_OVERLAPPEDWINDOW, false);
	m_hMainWnd = CreateWindow("D3DWndClassName", m_MainWndCaption.c_str(), 
		WS_OVERLAPPEDWINDOW, 100, 100, R.right, R.bottom, 
		0, 0, m_hAppInst, 0); 

	if( !m_hMainWnd )
	{
		MessageBox(0, "CreateWindow FAILED", 0, 0);
		PostQuitMessage(0);
	}

	ShowWindow(m_hMainWnd, SW_SHOW);
	UpdateWindow(m_hMainWnd);
	m_WindowRect = R;
}

void DirectX9App::InitDirect3D()
{
	//Create the IDirect3D9 object.
	m_D3DObject = Direct3DCreate9(D3D_SDK_VERSION);
	if( !m_D3DObject )
	{
		MessageBox(0, "Direct3DCreate9 FAILED", 0, 0);
		PostQuitMessage(0);
	}

	//Check for requested vertex processing and pure device.
	D3DCAPS9 caps;
	HR(m_D3DObject->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &caps));

	DWORD devBehaviorFlags = 0;

	if(caps.DevCaps & D3DCREATE_HARDWARE_VERTEXPROCESSING)
	{
		devBehaviorFlags |= D3DCREATE_HARDWARE_VERTEXPROCESSING;	
	}
	else
	{
		devBehaviorFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING; 
	}

	if(caps.DevCaps & D3DDEVCAPS_PUREDEVICE && devBehaviorFlags & D3DCREATE_HARDWARE_VERTEXPROCESSING)
	{
		devBehaviorFlags |= D3DCREATE_PUREDEVICE;	
	}

	//Fill out the D3DPRESENT_PARAMETERS structure.
	m_D3DPP.BackBufferWidth            = BASE_WINDOW_WIDTH; 
	m_D3DPP.BackBufferHeight           = BASE_WINDOW_HEIGHT;
	m_D3DPP.BackBufferFormat           = D3DFMT_UNKNOWN;
	m_D3DPP.BackBufferCount            = 1;
	m_D3DPP.MultiSampleType            = D3DMULTISAMPLE_NONE;
	m_D3DPP.MultiSampleQuality         = 0;
	m_D3DPP.SwapEffect                 = D3DSWAPEFFECT_COPY; 
	m_D3DPP.hDeviceWindow              = m_hMainWnd;
	m_D3DPP.Windowed                   = TRUE;
	m_D3DPP.EnableAutoDepthStencil     = TRUE; 
	m_D3DPP.AutoDepthStencilFormat     = D3DFMT_D16;
	m_D3DPP.Flags                      = D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;
	m_D3DPP.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
	m_D3DPP.PresentationInterval       = D3DPRESENT_INTERVAL_IMMEDIATE;
				
	//Create the device.
	HR(m_D3DObject->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, m_hMainWnd, devBehaviorFlags, &m_D3DPP, &m_D3DDevice)); 

	//Create sprite interface
	HR(D3DXCreateSprite(m_D3DDevice, &m_SpriteInterface));

	//Create the default font
	HR(D3DXCreateFont(gD3DDevice, 30, 0, FW_BOLD, 0, false, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, TEXT("Raavi"), &m_Font));

}
	
void DirectX9App::Shutdown()
{
	SAFE_RELEASE(m_Font);
	SAFE_RELEASE(m_SpriteInterface);
	SAFE_RELEASE(m_D3DDevice);
	SAFE_RELEASE(m_D3DObject);
}

LRESULT DirectX9App::MessageProc(UINT msg, WPARAM wParam, LPARAM lParam)
{
	// Is the application in a minimized or maximized state?
	static bool minOrMaxed = false;

	RECT clientRect = {0, 0, 0, 0};
	switch( msg )
	{

	//Window activate/deactivate
	case WM_ACTIVATE:
		if( LOWORD(wParam) == WA_INACTIVE )
			m_AppPaused = true;
		else
			m_AppPaused = false;
		return 0;


	//Window Resize
	case WM_SIZE:
		if( m_D3DDevice )
		{
			m_D3DPP.BackBufferWidth  = LOWORD(lParam);
			m_D3DPP.BackBufferHeight = HIWORD(lParam);

			if( wParam == SIZE_MINIMIZED )
			{
				m_AppPaused = true;
				minOrMaxed = true;
			}
			else if( wParam == SIZE_MAXIMIZED )
			{
				m_AppPaused = false;
				minOrMaxed = true;
				onLostDevice();
				HR(m_D3DDevice->Reset(&m_D3DPP));
				onResetDevice();
			}
			
			else if( wParam == SIZE_RESTORED )
			{
				m_AppPaused = false;

				// Are we restoring from a mimimized or maximized state, and in windowed mode?
				if( minOrMaxed && m_D3DPP.Windowed )
				{
					onLostDevice();
					HR(m_D3DDevice->Reset(&m_D3DPP));
					onResetDevice();
				}
				minOrMaxed = false;
			}
		}
		return 0;


	// Reset everything based on the new window dimensions.
	case WM_EXITSIZEMOVE:
		GetClientRect(m_hMainWnd, &clientRect);
		m_D3DPP.BackBufferWidth  = clientRect.right;
		m_D3DPP.BackBufferHeight = clientRect.bottom;
		onLostDevice();
		HR(m_D3DDevice->Reset(&m_D3DPP));
		onResetDevice();

		return 0;

	//Window close
	case WM_CLOSE:
		DestroyWindow(m_hMainWnd);
		return 0;

	// WM_DESTROY is sent when the window is being destroyed.
	case WM_DESTROY:
		Shutdown();
		PostQuitMessage(0);
		return 0;

	case WM_KEYDOWN:
		return 0;
	}
	return DefWindowProc(m_hMainWnd, msg, wParam, lParam);
}

//Application Entry Point
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE prevInstance, PSTR cmdLine, int showCmd)
{
	// Enable run-time memory check for debug builds.
	#if defined(DEBUG) | defined(_DEBUG)
		_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	#endif

	//Init Input
	gInput->Init(gApplication->GetMainWnd(), hInstance);

    return DirectX9App::Instance()->Run();
}

void DirectX9App::BeginFrame()
{
	// Clear the back buffer, call BeginScene()
	HR(gD3DDevice->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER , m_ClearColor, 1.0f, 0));
	HR(gD3DDevice->BeginScene());

	//Start sprite operations
	HR(gSprite->Begin(D3DXSPRITE_ALPHABLEND | D3DXSPRITE_SORT_DEPTH_FRONTTOBACK));

}

void DirectX9App::EndFrame()
{
	//End sprite operations
	HR(gSprite->End());

	// EndScene, and Present the back buffer to the display buffer
	HR(gD3DDevice->EndScene());
	HR(gD3DDevice->Present(0, 0, 0, 0));
}

IDirect3DVertexDeclaration9* VertexPNT::Decl;

VertexPNT* VertexPNT::Instance()	
{
	static VertexPNT instance;
	return &instance;
}
		