//////////////////////////////////////////////////////////////////////////////
//	DirectX9App.h
//	Tylor Emmett - 2/25/2014
//	Provides a Win32 window, Direct3D9 Graphics, and houses the main game loop
//	Singleton Class.
//	General access via the Instance() method, defined as gApplication in Defines.h
///////////////////////////////////////////////////////////////////////////////

#pragma once
#include <d3d9.h>
#include <d3dx9.h>
#include <dxerr.h>
#include "Defines.h"
#include <string>
#include <vector>


#define APP_TITLE "GSP 361 - Tank Gauntlet"
#define BASE_WINDOW_WIDTH 1280
#define BASE_WINDOW_HEIGHT 1024

class DirectX9App
{
private:
	/////////////////////////////////////////////////////////////////////////
	//	DirectX/Windows API Members
	/////////////////////////////////////////////////////////////////////////
	std::string				m_MainWndCaption;
	HINSTANCE				m_hAppInst;
	HWND					m_hMainWnd;
	IDirect3D9*				m_D3DObject;
	IDirect3DDevice9*		m_D3DDevice;
	D3DPRESENT_PARAMETERS	m_D3DPP;
	D3DXCOLOR				m_ClearColor;
	bool					m_AppPaused;
	ID3DXSprite*			m_SpriteInterface;
	ID3DXFont*				m_Font;
	RECT					m_WindowRect;

	DirectX9App();
	DirectX9App(DirectX9App&){}
	void operator=(DirectX9App&){}
public:

	~DirectX9App();
	
	/////////////////////////////////////////////////////////////////////////
	//	Utility Functions
	/////////////////////////////////////////////////////////////////////////

	//Changes between windowed and fullscreen mode.
	void EnableFullScreenMode(bool enable);

	/////////////////////////////////////////////////////////////////////////
	//	Accesors
	/////////////////////////////////////////////////////////////////////////

	//Accessor for the handle to the application instance
	HINSTANCE GetAppInst(){ return m_hAppInst; }
	//Accessor for the handle to the window
	HWND GetMainWnd(){ return m_hMainWnd; }
	//Accessor for the main D3D Device
	IDirect3DDevice9* GetDevice(){ return m_D3DDevice; }
	//Accessor for the Sprite interface
	ID3DXSprite* GetSprite(){ return m_SpriteInterface; }
	//Accessor for the default font
	ID3DXFont* GetDefaultFont(){ return m_Font; }
	//Handle to the static instance of this class
	static DirectX9App* Instance();
	//Accessor for the current window size
	D3DXVECTOR2 GetWindowSize(){return D3DXVECTOR2((float)m_D3DPP.BackBufferWidth, (float)m_D3DPP.BackBufferHeight);}
	//Accessor for window rect
	RECT GetWindowRect(){return m_WindowRect;}

	/////////////////////////////////////////////////////////////////////////
	//	Direct3D Reset Functions
	/////////////////////////////////////////////////////////////////////////
	
	//Processed on lost device. Destory non DirectX managed objects here.
	void onLostDevice();
	//Processed on reset device. Reset non DirectX managed objects here.
	void onResetDevice();
	//Tests for a lost device.
	bool isDeviceLost();

	/////////////////////////////////////////////////////////////////////////
	//	Direct3D/Windows API Control Functions
	/////////////////////////////////////////////////////////////////////////
	
	//Check for device capabilities
	bool CheckDeviceCaps();
	//Setup main window
	void InitMainWindow();
	//Startup Direct3D
	void InitDirect3D();
	//Window message loop
	int Run();
	//Shutdown Application
	void Shutdown();
	//Message Processor
	LRESULT MessageProc(UINT msg, WPARAM wParam, LPARAM lParam);
	//Start/Stop functions for rendering.
	void BeginFrame();
	void EndFrame();
	
};

//Vertex Struct for a Position, Normal and Tex Coord Vertex
	struct VertexPNT
	{
	private:
		VertexPNT()
			:pos(0.0f, 0.0f, 0.0f),
			normal(0.0f, 0.0f, 0.0f),
			tex0(0.0f, 0.0f)
		{
			D3DVERTEXELEMENT9 VertexPNTElements[] = 
			{
			{0, 0,  D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
			{0, 12, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL, 0},
			{0, 24, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
			D3DDECL_END()
			};	
			HR(gD3DDevice->CreateVertexDeclaration(VertexPNTElements, &Decl));
		}
	public:
		~VertexPNT()
		{
			SAFE_RELEASE(Decl);	
		}

		D3DXVECTOR3 pos;
		D3DXVECTOR3 normal;
		D3DXVECTOR2 tex0;

		static VertexPNT* Instance();
	
		static IDirect3DVertexDeclaration9* Decl;
	};

	
