#include "Projectile.h"
#include "DirectX9App.h"
#include "ActorManager.h"
#include "PlayerTank.h"
#include "AITank.h"
#include "Camera.h"
#include "AssetManager.h"
#include <sstream>
#include "FMODAudio.h"

Projectile::Projectile(int ID, int owner, float flightTime, int damage, float scale):Actor(ID), m_Owner(owner), m_Damage(damage), m_FlightTime(flightTime), m_Scale(scale)
{
	ZeroMemory(&m_Material, sizeof(m_Material));

	m_Material.Ambient = D3DXCOLOR(0.2f, 0.2f, 0.2f, 1.0f);		// Ambient color reflected
	m_Material.Diffuse = D3DXCOLOR(1.0f, 1.0f, 1.0f, 1.0f);		// Diffuse color reflected
	m_Material.Emissive = D3DXCOLOR(0.0f, 0.0f, 0.0f, 1.0f);		// Emissive color reflected
	m_Material.Specular = D3DXCOLOR(0.8f, 0.8f, 0.8f, 0.8f);		// Specular
	m_Material.Power = 50000.0f;									// Specular highlight intensity
	
	m_hTech = gAssets->GetShader("ProjectB.fx")->GetTechniqueByName("tech0");
}


void Projectile::Update(float dt)
{
	m_Physics.Update(dt);

	//Destory self if flight time elapses
	m_FlightTime -= dt;
	if(m_FlightTime <= 0.0f)
	{
		gMessenger->AddMessage(m_ID, m_ID, msg_Destroy, 0.0f);
	}
	//If player owned projectile, process contacts between it and the 2 enemies.
	if(m_Owner == gActors->GetPlayer()->GetID())
	{
		AITank* target1 = (AITank*)gActors->GetActiveAITank();
		Actor* target2 = gActors->GetActiveCannon();
		if(target1 != nullptr)
		{
			float distance1 = D3DXVec3Length(&(m_Physics.GetPosition() - target1->GetTank()->GetPhysics()->GetPosition()));
			//Check against AI Tank
			if(distance1 <= target1->GetTank()->GetPhysics()->GetContactRadius())
			{
				target1->AdjustHealth(-m_Damage);
				gMessenger->AddMessage(m_ID, m_ID, msg_Destroy, 0.0f);
			}
		}
	}
	//Otherwise, process contacts between it and the player.
	else
	{
		PlayerTank* target = (PlayerTank*)gActors->GetPlayer();
		if(target == nullptr)
			return;
		float distance = D3DXVec3Length(&(m_Physics.GetPosition() - target->GetTank()->GetPhysics()->GetPosition()));
		if(distance <= target->GetTank()->GetPhysics()->GetContactRadius())
		{
			target->AdjustHealth(-m_Damage);
			gAudio->PlaySoundFX("Machinegun_Hit.wav");
			gMessenger->AddMessage(m_ID, m_ID, msg_Destroy, 0.0f);
		}
	}

}

void Projectile::Render()
{
	D3DXMATRIX transMat, rotMat, scaleMat, worldMat, WIT;
	D3DXMatrixIdentity(&transMat);
	D3DXMatrixIdentity(&scaleMat);
	D3DXMatrixIdentity(&rotMat);
	D3DXMatrixIdentity(&worldMat);
	
	gAssets->GetShader("ProjectB.fx")->SetTechnique(m_hTech);

	//Get light and camera data
	D3DLIGHT9 mainLight = gActors->GetMainLight();
	D3DXMATRIX viewMat = ((Camera*)gActors->GetActiveCamera())->GetView();
	D3DXMATRIX projMat = ((Camera*)gActors->GetActiveCamera())->GetProjection();
	D3DXVECTOR3 camPosition = ((Camera*)gActors->GetActiveCamera())->GetPosition();

	gAssets->GetShader("ProjectB.fx")->SetFloatArray("eyePos", (float*)camPosition, 3);
	gAssets->GetShader("ProjectB.fx")->SetFloatArray("lightPos", (float*)&mainLight.Position, 3);
	gAssets->GetShader("ProjectB.fx")->SetFloatArray("ambientLight", (float*)&mainLight.Ambient, 3);
	gAssets->GetShader("ProjectB.fx")->SetFloatArray("specularLight", (float*)&mainLight.Specular, 3);
	gAssets->GetShader("ProjectB.fx")->SetFloatArray("diffuseLight", (float*)&mainLight.Diffuse, 3);
	gAssets->GetShader("ProjectB.fx")->SetFloatArray("lightAttenuation", (float*)&D3DXVECTOR3(mainLight.Attenuation0, mainLight.Attenuation1, mainLight.Attenuation2), 3);

	
	// Begin Effect
	UINT numPasses = 0;
	gAssets->GetShader("ProjectB.fx")->Begin(&numPasses, 0);

	float yawRot = atan2(m_Physics.GetVelocity().x, m_Physics.GetVelocity().z);

	for(UINT i = 0; i < numPasses; ++i)
	{
		gAssets->GetShader("ProjectB.fx")->BeginPass(i);
	
	
		// Calculate Matrix Transform
		D3DXMatrixScaling(&scaleMat, m_Scale, m_Scale, m_Scale);			// Scaling
		D3DXMatrixRotationYawPitchRoll(&rotMat, yawRot, 0.0f, 0.0f); // Rotation on Yaw, Pitch, and Roll
		D3DXMatrixTranslation(&transMat, m_Physics.GetPosition().x, m_Physics.GetPosition().y, m_Physics.GetPosition().z);		// Translation
		D3DXMatrixMultiply(&scaleMat, &scaleMat, &rotMat);		// Multiply scale and rotation, store in scale
		D3DXMatrixMultiply(&worldMat, &scaleMat, &transMat);	// Multiply scale and translation, store in world
		
		D3DXMatrixInverse(&WIT, 0, &worldMat);
		D3DXMatrixTranspose(&WIT, &WIT);
		gAssets->GetShader("ProjectB.fx")->SetMatrix("worldViewProjMat", &(worldMat * viewMat * projMat));
		gAssets->GetShader("ProjectB.fx")->SetMatrix("worldInverseTransposeMat", &WIT);
		gAssets->GetShader("ProjectB.fx")->SetMatrix("worldMat", &worldMat);
		gAssets->GetShader("ProjectB.fx")->SetValue("ambientMaterial", &m_Material.Ambient, sizeof(D3DXCOLOR));
		gAssets->GetShader("ProjectB.fx")->SetValue("diffuseMaterial", &m_Material.Diffuse, sizeof(D3DXCOLOR));
		gAssets->GetShader("ProjectB.fx")->SetValue("specularMaterial", &m_Material.Specular, sizeof(D3DXCOLOR));
		gAssets->GetShader("ProjectB.fx")->SetFloat("specularPower", m_Material.Power);
		gAssets->GetShader("ProjectB.fx")->SetTexture("tex", gAssets->GetTexture("GreyTank.png"));

		gAssets->GetShader("ProjectB.fx")->CommitChanges();
		
		gAssets->GetMesh("Bullet.x")->DrawSubset(0);
		
		gAssets->GetShader("ProjectB.fx")->EndPass();
	}

	gAssets->GetShader("ProjectB.fx")->End();
}

void Projectile::ReceiveMessage(Telegram incoming)
{
	switch(incoming.msg)
	{
	case msg_Destroy:
		if(incoming.sender = m_ID)
			gActors->RemoveActor(m_ID);
	}
}